@@ -1,24 +1,23 @@
 package Algorithm::MinPerfHashTwoLevel;
 use strict;
 use warnings;
-our $VERSION = '0.10';
-our $DEFAULT_VARIANT = 2;
+our $VERSION = '0.14';
+our $DEFAULT_VARIANT = 3;
 
 use Exporter qw(import);
+use Carp ();
 
 no warnings "portable";
 my %constant;
 BEGIN {
     %constant= (
-        NOT_UTF8 => 0,
-         IS_UTF8 => 1,
-        WAS_UTF8 => 2,
-
-        UINT64_MAX => 0xFFFFFFFFFFFFFFFF,
-        UINT32_MAX => 0xFFFFFFFF,
-        INT32_MAX  => 0x7FFFFFFF,
-        UINT16_MAX => 0xFFFF,
-        UINT8_MAX  => 0xFF,
+        MPH_F_FILTER_UNDEF          =>  (1<<0),
+        MPH_F_DETERMINISTIC         =>  (1<<1),
+       #MPH_F_NO_DEDUPE             =>  (1<<2),
+       #MPH_F_VALIDATE              =>  (1<<3),
+        MAX_VARIANT                 =>  3,
+        STADTX_HASH_SEED_BYTES      => 16,
+        STADTX_HASH_STATE_BYTES     => 32,
     );
 }
 use constant \%constant;
@@ -26,13 +25,13 @@ use constant \%constant;
 our %EXPORT_TAGS = (
     'all' => [
         '$DEFAULT_VARIANT',
+        'MAX_VARIANT',
         qw(
             seed_state
             hash_with_state
         ), sort keys %constant
     ],
-    'utf8_flags' => [ grep /UTF8/, sort keys %constant],
-    'uint_max'   => [ grep /_MAX/, sort keys %constant]
+    'flags' => [ sort grep /MPH_F_/, keys %constant ],
 );
 
 our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );
@@ -49,84 +48,125 @@ use Carp ();
 
 sub new {
     my ($class,%opts)= @_;
+    my $seed= delete($opts{seed});
+    delete($opts{state}) and warn "ignoring 'state' parameter";
+
     my $o= bless \%opts, $class;
-    $o->{state} = seed_state($o->{seed})
-        if $o->{seed};
+
+    $o->set_seed($seed) if $seed;
+
     $o->{variant}= $DEFAULT_VARIANT unless defined $o->{variant};
     $o->{variant}= int(0+$o->{variant});
-    $o->{compute_flags}=0;
-    $o->{compute_flags} += 1 if delete $o->{filter_undef};
-    $o->{compute_flags} += 2 if delete $o->{deterministic};
-    die "Unknown variant '$o->{variant}' in constructor new(), max known is 2"
-        if ($o->{variant} > 2);
+
+    $o->{compute_flags} ||= 0;
+    $o->{compute_flags} += MPH_F_FILTER_UNDEF
+        if delete $o->{filter_undef};
+    $o->{compute_flags} += MPH_F_DETERMINISTIC
+        if delete $o->{deterministic} or delete $o->{canonical};
+
+    die "Unknown variant '$o->{variant}' in constructor new(), max known is "
+        . MAX_VARIANT . " default is " . $DEFAULT_VARIANT
+        if $o->{variant} > MAX_VARIANT;
+
     return $o;
 }
 
-# find a suitable initial seed for
-sub _compute_first_level {
-    my ($self)= @_;
-    my $source_hash= $self->{source_hash};
+sub compute {
+    my ($self, $source_hash)= @_;
+    if ($source_hash) {
+        $self->{source_hash}= $source_hash;
+    } else {
+        $source_hash= $self->{source_hash};
+    }
     my $debug= $self->{debug} ||= 0;
 
-    printf"computing first level hash information for %d keys\n",
-        0+keys %$source_hash
-        if $debug;
+    # reuse the constructor seed.
+    $self->_seed($self->{constructor_seed});
 
     # find the number of keys we have to deal with
-    my $n= $self->{n}= 0+keys %$source_hash;
-    my $max_tries= $self->{max_tries} || 100;
-    my $min_tries= $self->{min_tries} || 1;
+    my $max_tries= $self->{max_tries} || 10;
+    my @failed_seeds;
+    $self->{failed_seeds}= \@failed_seeds;
 
-
-    # Find the base seed and build a map of the keys to the buckets that they will reside in
-    SEED1:
     for my $counter ( 1 .. $max_tries ) {
-        if (!defined $self->{seed}) {
-            $self->{seed}= join "", map { chr(rand 256) } 1 .. 16;
-        }
-        if (!defined $self->{state}) {
-            $self->{state} = seed_state($self->{seed});
-        }
-        my $buckets= $self->_compute_first_level_inner();
-        if ($buckets) {
-            return $buckets;
-        } else {
-            print "seed failed, trying new seed\n" if $debug;
-            delete $self->{seed};
-            delete $self->{state};
+        my $seed= $self->get_seed; # ensure we have a seed set up (must be called before compute_xs)
+        my $state= $self->get_state; # ensure we have a state set up (must be called before compute_xs)
+        delete $self->{buckets};
+
+        printf "MPH2L compute attempt #%2d/%2d for hash with %6d keys - using seed: %s (state: %s)\n",
+            $counter,
+            $max_tries,
+            0+keys(%$source_hash),
+            unpack("H*",$seed),
+            unpack("H*",$state),
+            if $debug;
+
+        my $bad_idx= compute_xs($self)
+            or return $self->{buckets};
+
+        push @failed_seeds, $seed;
+        if ($counter < $max_tries) {
+            $self->re_seed();
         }
     }
-    Carp::confess("This is unexpected. We tried $max_tries times to find a seed with the appropriate properties, and we failed.\n",
-        join " ", sort keys(%$source_hash));
+
+    if ( $max_tries == 1 ) {
+        die sprintf "Failed to compute minimal perfect hash using seed %s", unpack "H*", $failed_seeds[0];
+    } else {
+        Carp::confess(
+            sprintf "Failed to compute minimal perfect hash after %d tries. Seeds tried: %s",
+                $max_tries, join(" ", map { unpack "H*", $_ } @failed_seeds)
+        );
+    }
+    # NOT REACHED.
 }
 
-sub _compute_first_level_inner {
+sub re_seed {
+    my $self= shift;
+
+    my $source= $self->get_state();
+    return $self->_seed(substr($source,0,STADTX_HASH_SEED_BYTES) ^ substr($source,STADTX_HASH_SEED_BYTES));
+}
+
+
+sub _seed {
+    my $self= shift;
+    if (@_) {
+        my $seed= shift;
+        Carp::confess(sprintf "Seed should be undef, or a string exactly %d bytes long, not %d bytes",
+            STADTX_HASH_SEED_BYTES,length($seed))
+            if defined($seed) and length($seed) != 16;
+        $self->{seed}= $seed;
+        delete $self->{state};
+    }
+    if ( !defined $self->{seed} ) {
+                       #1234567812345678
+        $self->{seed}= "MinPerfHash2Levl";
+        delete $self->{state};
+    }
+    return $self->{seed};
+}
+
+sub set_seed {
+    my ($self,$value)= @_;
+    my $seed= $self->_seed($value);
+    return $self->{constructor_seed}= $seed;
+}
+
+sub get_seed {
     my ($self)= @_;
-    my $debug= $self->{debug};
-
-    printf "checking seed %s => state: %s\n", 
-        unpack("H*",$self->{seed}), 
-        unpack("H*",$self->{state}), 
-        if $debug;
-
-    my $bad_idx= compute_xs($self);
-    if ($bad_idx) {
-        printf " Index '%d' not solved, new seed required.\n", $bad_idx-1 if $debug;
-        return undef;
-    }
-
-    return $self->{buckets};
+    return $self->_seed();
 }
 
-sub compute {
-    my ($self,$source_hash)= @_;
-    $self->{source_hash}= $source_hash if $source_hash;
 
-    return $self->_compute_first_level();
+sub get_state {
+    my $self= shift;
+    return $self->{state} ||= seed_state($self->_seed());
 }
 
-sub state {
-    return $_[0]->{state};
+sub get_failed_seeds {
+    my $self= shift;
+    return @{$self->{failed_seeds}||[]};
 }
 
 1;
@@ -178,7 +218,7 @@ In essence this module performs the task of computing the xor_val for
 each bucket such that the idx2 for every element is unique, it does it in C/XS so
 that it is fast.
 
-The INTHASH() function used depends by variant, with variant 2 it is:
+The INTHASH() function used depends by variant, with variant 2 and later it is:
 
     x = ((x >> 16) ^ x) * 0x45d9f3b;
     x = ((x >> 16) ^ x) * 0x45d9f3b;
@@ -187,7 +227,7 @@ The INTHASH() function used depends by variant, with variant 2 it is:
 which is just a simple 32 bit integer hash function I found at
 https://stackoverflow.com/a/12996028, but any decent reversible
 integer hash function would do. For variant 0 and 1 it is the identity
-function. The default is variant 2.
+function. The default is variant 3.
 
 *NOTE* in Perl a given string may have differing binary representations
 if it is encoded as utf8 or not. This module uses the same conventions
